/* -*- c-basic-offset:4; indent-tabs-mode: nil -*- */
/* ====================================================================
 * Copyright (c) 2022 David Huggins-Daines.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */
/**
 * @file endpointer.h
 * @brief VAD-based endpointer
 */

#ifndef __ENDPOINTER_H__
#define __ENDPOINTER_H__

#ifdef __cplusplus
extern "C" {
#endif
#if 0
}
#endif

#include <soundswallower/prim_type.h>
#include <soundswallower/vad.h>

/**
 * @struct endpointer_t
 * @brief Simple voice activity detection based endpointing
 */
typedef struct endpointer_s endpointer_t;

/**
 * Default window in seconds of audio to use for speech start/end decision.
 */
#define ENDPOINTER_DEFAULT_WINDOW 0.3
/**
 * Default ratio of frames in window to trigger start/end decision.
 */
#define ENDPOINTER_DEFAULT_RATIO 0.9

/**
 * Initialize endpointing.
 *
 * @memberof endpointer_t
 * @param window Seconds of audio to use in speech start/end decision,
 *               or 0 to use the default (ENDPOINTER_DEFAULT_WINDOW).
 * @param ratio Ratio of frames needed to trigger start/end decision,
 *              or 0 for the default (ENDPOINTER_DEFAULT_RATIO).
 * @param mode "Aggressiveness" of voice activity detection.  Stricter
 *             values (see vad_mode_t) are less likely to
 *             misclassify non-speech as speech.
 * @param sample_rate Sampling rate of input, or 0 for default (which can
 *                    be obtained with vad_sample_rate()).  Only 8000,
 *                    16000, 32000, 48000 are directly supported, others
 *                    will use the closest supported rate (within reason).
 *                    Note that this means that the actual frame length
 *                    may not be exactly the one requested, so you must
 *                    always use the one returned by
 *                    endpointer_frame_size()
 *                    (in samples) or endpointer_frame_length() (in
 *                    seconds).
 * @param frame_length Requested frame length in seconds, or 0.0 for the
 *                     default.  Only 0.01, 0.02, 0.03 currently supported.
 *                     **Actual frame length may be different, you must
 *                     always use endpointer_frame_length() to obtain it.**
 * @return Endpointer object or NULL on failure (invalid parameter for
 * instance).
 */
endpointer_t *endpointer_init(double window,
                              double ratio,
                              vad_mode_t mode,
                              int sample_rate, double frame_length);

/**
 * Retain a pointer to endpointer
 *
 * @memberof endpointer_t
 * @param ep Endpointer.
 * @return Endpointer with incremented reference count.
 */
endpointer_t *endpointer_retain(endpointer_t *ep);

/**
 * Release a pointer to endpointer.
 *
 * @memberof endpointer_t
 * @param ep Endpointer
 * @return New reference count (0 if freed).
 */
int endpointer_free(endpointer_t *ep);

/**
 * Get the voice activity detector used by the endpointer.
 *
 * @memberof endpointer_t
 * @return VAD object. The endpointer retains ownership of this
 * object, so you must use vad_retain() if you wish to use it
 * outside of the lifetime of the endpointer.
 */
vad_t *endpointer_vad(endpointer_t *ep);

/**
 * Get the frame size (in samples) consumed by the endpointer.
 *
 * Multiply this by 2 to get the size of the frame buffer required.
 */
#define endpointer_frame_size(ep) vad_frame_size(endpointer_vad(ep))

/**
 * Get the frame length (in seconds) consumed by the endpointer.
 */
#define endpointer_frame_length(ep) vad_frame_length(endpointer_vad(ep))

/**
 * Get the sample rate required by the endpointer.
 */
#define endpointer_sample_rate(ep) vad_sample_rate(endpointer_vad(ep))

/**
 * Process a frame of audio, returning a frame if in a speech region.
 *
 * Note that the endpointer is *not* thread-safe.  You must call all
 * endpointer functions from the same thread.
 *
 * @memberof endpointer_t
 * @param ep Endpointer.
 * @param frame Frame of data, must contain endpointer_frame_size()
 *              samples.
 * @return NULL if no speech available, or pointer to a frame of
 *         endpointer_frame_size() samples (no more and no less).
 */
const int16 *endpointer_process(endpointer_t *ep,
                                const int16 *frame);

/**
 * Process remaining samples at end of stream.
 *
 * Note that the endpointer is *not* thread-safe.  You must call all
 * endpointer functions from the same thread.
 *
 * @memberof endpointer_t
 * @param ep Endpointer.
 * @param frame Frame of data, must contain endpointer_frame_size()
 *              samples or less.
 * @param nsamp: Number of samples in frame.
 * @param out_nsamp: Output, number of samples available.
 * @return Pointer to available samples, or NULL if none available.
 */
const int16 *endpointer_end_stream(endpointer_t *ep,
                                   const int16 *frame,
                                   size_t nsamp,
                                   size_t *out_nsamp);

/**
 * Get the current state (speech/not-speech) of the endpointer.
 *
 * This function can be used to detect speech/non-speech transitions.
 * If it returns 0, and a subsequent call to endpointer_process()
 * returns non-NULL, this indicates a transition to speech.
 * Conversely, if endpointer_process() returns non-NULL and a
 * subsequent call to this function returns 0, this indicates a
 * transition to non-speech.
 *
 * @memberof endpointer_t
 * @param ep Endpointer.
 * @return non-zero if in a speech segment after processing the last
 *         frame of data.
 */
int endpointer_in_speech(endpointer_t *ep);

/**
 * Get the start time of the last speech segment.
 * @memberof endpointer_t
 */
double endpointer_speech_start(endpointer_t *ep);

/**
 * Get the end time of the last speech segment
 * @memberof endpointer_t
 */
double endpointer_speech_end(endpointer_t *ep);

#ifdef __cplusplus
}
#endif

#endif /* __ENDPOINTER_H__ */
